//
// UNR2DE
//
// Unreal to Deus Ex mesh converter.  Converts old style Unreal
// and Unreal Tournament _a.3d files into the higher precision
// format that Deus Ex expects.
//
// Takes 1 argument: the "base" filename of a pair of Unreal mesh
// filenames.  For example, if "RoboKitty" is passed in, the
// program will look at the "RoboKitty_a.3d" and "RoboKitty_d.3d"
// files.  It *will* overwrite the "_a.3d" file.
//
// Written by Steve Tack, 10/15/00
//
// Revision History
// ----------------
// v1.1 10/26/00 Changed datatype of NumFrames, FrameSize, NumPolys,
//               and NumVertices to "unsigned short" instead of just
//               "short" to allow for very large meshes to be processed.
//               (frame sizes over 32,000 were going negative).
// v1.2 08/25/01 Added functionality for "MilkShape compensation" by 
//               accepting a "-ms" option.  When that option is specified,
//               the X-axis is flipped, the polygons are flipped, and the
//               UV coordinates are flipped.  This feature allows models
//		 to convert properly when using MilkShape's Unreal/UT
//		 exporter. (you no longer have to do Mirror Left <-> Right
//		 in MilkShape)

#define WIN32_LEAN_AND_MEAN

#include <stdio.h>
#include <string>

//-----------------------------------------------------------------------

void main( int argc, char* argv[] )
{

	printf("\nUnreal to Deus Ex mesh converter, v1.2");
	printf("\nWritten by Steve Tack\n");

	// Display info if no command-line arguments passed in
	if(argc == 1) {

		printf("\nunr2de will convert an Unreal or Unreal Tournament mesh file");
		printf("\ninto a format suitable for Deus Ex. This allows the use of");
		printf("\nexisting Unreal tools that do not work with Deus Ex.\n");
		printf("\nUsage: unr2de <part of filename before the \"_\">\n");
		printf("\nFor example:");
		printf("\nunr2de BunnyDroid");
		printf("\nwill convert a file named BunnyDroid_a.3d\n");
		printf("\nThe -ms option will compensate for models exported from MilkShape.");
		printf("\nFor example:");
		printf("\nunr2de -ms BunnyDroid\n");
		printf("\nYou will need to have your .3d files in the same directory.");
		printf("\nThis program *will* overwrite your _a.3d file, so make a backup");
		printf("\nof it if necessary.  Your _d.3d file will also be overwritten");
		printf("\nwhen using the -ms option.\n");
		return;
	}

    // Parse options
    int CurArg;
	int FilenameArg = 0;
	int FixMilkShape = 0;
	int HighResExperiment = 0;
    for( CurArg = 1; CurArg < argc; ++CurArg ) {
        if( *argv[ CurArg ] == '-' ) {

            if( !stricmp( argv[ CurArg ], "-ms" ) ) {
                FixMilkShape = 1;

            } else if (!stricmp( argv[ CurArg ], "-h" ) ) {
                HighResExperiment = 1;
            }
        } else { 
	
			FilenameArg = CurArg;	// This arg should be the filename (last arg)
            break;  // done with options
        }
    }

	// Construct the "_a.3d" and "_d.3d" filenames.  Be sure to allocate an
	// extra 5 characters for the characters at the end.
	char *AnimFileName = new char[strlen(argv[FilenameArg]) + 6];
	strcpy(AnimFileName, argv[FilenameArg]);
	strcat(AnimFileName, "_a.3d");

	char *DataFileName = new char[strlen(argv[FilenameArg]) + 6];
	strcpy(DataFileName, argv[FilenameArg]);
	strcat(DataFileName, "_d.3d");

	// Open the _a.3d file (binary mode)
	FILE* fp = fopen(AnimFileName, "rb");
	if (fp == 0) {
		printf("\n\nCan't open file %s !\n", AnimFileName);
		return;
	}

	unsigned short NumFrames, FrameSize;
	unsigned short NumPolys, NumVertices;
	unsigned int BytesPerVertex;

	// Read in number of animation frames
	if (fread(&NumFrames, sizeof(NumFrames), 1, fp) != 1) {
		printf("\n\nProblem reading data file: could not read number of frames !\n");
		return;
	}

	// Time to rock and roll
	printf("\nConverting %s...\n", AnimFileName);

	// Spit out a message if we're in "MilkShape Fix" mode
	if (FixMilkShape)
		printf("\n* MilkShape Fix Mode *\n");

	// Spit out a message if we're in experimental "hi res" mode
	if (HighResExperiment)
		printf("\n* Hi Res Experimental Mode *\n");

	// Display number of animation frames, just because we have it
	// and it's nice to tell the user something's happening
	if (NumFrames == 1)
		printf("\n  Single mesh (no animation)");
	else if (NumFrames > 1)
		printf("\n  %d animation frames", NumFrames);
	else if (NumFrames < 1) {
		printf("\n\nNumber of animation frames invalid !\n");
		return;
	}

	// Read in "frame size" (how many bytes each frame of animation is)
	// For Unreal files, this is 4 bytes * the number of vertices. Don't 
	// display this number to the user, cause he doesn't care
	if (fread(&FrameSize, sizeof(FrameSize), 1, fp) != 1) {
		printf("\n\nProblem reading data file: could not read frame size !\n");
		return;
	}

	// Struct for polys for output to _d.3d if the "-ms"
	// option is used
	typedef struct DeusExPolygonStruct {
		unsigned short PolyIdx1;
		unsigned short PolyIdx2;
		unsigned short PolyIdx3;
		unsigned char TriangleType;
		unsigned char Color;
		unsigned char Ucoord1;
		unsigned char Vcoord1;
		unsigned char Ucoord2;
		unsigned char Vcoord2;
		unsigned char Ucoord3;
		unsigned char Vcoord3;
		unsigned char TextureNum;
		unsigned char dummy;
	} DeusExPolygonType;

	DeusExPolygonType *DeusExPolygon;
	long dummy_long;

	// Now temporarily open the "_d.3d" file to grab the number of vertices.
	// The only reason we're doing this is to verify that the input file
	// has not already been converted.  (frame size / number of verticies
	// should be 4 for an Unreal file)
	FILE* fp_datafile = fopen( DataFileName, "rb");
	if (fp_datafile == 0) {
		printf("\n\nCan't open file %s !\n", DataFileName);
		return;
	}

	// Read number of polygons from _d.3d file
	if (fread(&NumPolys, sizeof(NumPolys), 1, fp_datafile) != 1) {
		printf("\n\nProblem reading data file: could not read number of polygons !\n");
		return;
	}
	else
		printf("\n  %d polygons", NumPolys);

	if (NumPolys < 1) {
		printf("\n\nNumber of polygons invalid !");
		return;
	}

	// Read number of vertices from _d.3d file
	if (fread(&NumVertices, sizeof(NumVertices), 1, fp_datafile) != 1) {
		printf("\n\nProblem reading data file: could not read number of vertices !\n");
		return;
	}
	else
		printf("\n  %d vertices", NumVertices);

	if (NumVertices < 1) {
		printf("\n\nNumber of vertices invalid !");
		return;
	}

	// If the -ms option is being used, read in all of the polys
	if (FixMilkShape) {
		// Read the block of 44 dummy bytes
		// A long is 4 bytes, so read in 11 of 'em
		dummy_long = 0;
		for (int i = 0; i < 11; i++)
			fread(&dummy_long, 4, 1, fp_datafile);

		// Allocate memory for the polygon data
		DeusExPolygon = (DeusExPolygonType *)calloc( NumPolys, sizeof(DeusExPolygonType) );

		// Read in polys and poly-related flags, UV coords, and texture associations
		for (i = 0; i < NumPolys; i++) {
			fread(&DeusExPolygon[i].PolyIdx1, 2, 1, fp_datafile);	// Index to 1st polygon vertex
			fread(&DeusExPolygon[i].PolyIdx2, 2, 1, fp_datafile);	// Index to 1st polygon vertex
			fread(&DeusExPolygon[i].PolyIdx3, 2, 1, fp_datafile);	// Index to 1st polygon vertex
			fread(&DeusExPolygon[i].TriangleType, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].Color, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].Ucoord1, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].Vcoord1, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].Ucoord2, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].Vcoord2, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].Ucoord3, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].Vcoord3, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].TextureNum, 1, 1, fp_datafile);
			fread(&DeusExPolygon[i].dummy, 1, 1, fp_datafile);
		}
	}

	// We're done with the "_d.3d" file, so close up
	fclose(fp_datafile);

	// Make sure we've got an Unreal file that hasn't already
	// been converted (Unreal files are 4 bytes per vertex,
	// Deus Ex files are 8 bytes per vertex)
	BytesPerVertex = (FrameSize / NumVertices);

	if (BytesPerVertex == 8) {
		printf("\n\nMesh file has already been converted !\n");
		return;
	}
	else if (BytesPerVertex != 4) {
		printf("\n\nCound not determine vertex format !\n");
		return;
	}

	//
	// Start reading in the vertex coordinates and
	// stick 'em in an array.
	//
	int *VertexCoord = new int[sizeof(int) * NumVertices * NumFrames];
	int arrcnt = 0;
	int i = 0;
	short minX = 9999, maxX = -9999;
	short minY = 9999, maxY = -9999;
	short minZ = 9999, maxZ = -9999;

	while (fread(&VertexCoord[arrcnt], sizeof(VertexCoord), 1, fp) == 1) 
		arrcnt++;

	// Close the "_a.3d" file (no more reading needed)
	fclose(fp);

	// Open up the "_a.3d" for writing (binary mode).
	FILE* fp_write = fopen( AnimFileName, "wb");
	if (fp_write == 0) {
		printf("\n\nCan't open file %s for writing !\n", AnimFileName);
		delete [] VertexCoord;
		return;
	}

	// Write number of animation frames (this will not change from
	// original value)
	if (fwrite(&NumFrames, sizeof(NumFrames), 1, fp_write) != 1) { 
		printf("\n\nProblem writing to file %s !", AnimFileName);
		delete [] VertexCoord;
		return;
	}

	// Write new frame size (double the old one, since we're going
	// from 4 bytes per vertex to 8)
	FrameSize *= 2;
	if (fwrite(&FrameSize, sizeof(FrameSize), 1, fp_write) != 1) { 
		printf("\n\nProblem writing to file %s !\n", AnimFileName);
		delete [] VertexCoord;
		return;
	}

	struct DeusExVertex { 
        short x;
        short y;
        short z;
        short dummy;
    } st;

	memset( &st, '\0', sizeof(st) );
	st.dummy = 0;

	//
	// Convert the vertices and write 'em out!
	//
	for(i = 0; i < arrcnt; i++) {
	
		// Calculate x, y, and z values from the compressed Unreal format
		st.x = (VertexCoord[i] & 0x7FF) / 8;

		if (st.x > 128 )
			st.x -= 256;

		// If the "-ms" option is set, flip the X axis. (this is confusing,
		// but normally, the st.x value would be flipped, here, so if the
		// -ms option is used, we *don't* negate the value)  This is
		// to compensate for the MilkShape Unreal exporter (version 1.4.1 anyway)
		// The polygons will also be flipped, as well as the UV coords
		if (!FixMilkShape) 
			st.x = -st.x;

		st.y = ((VertexCoord[i] >> 11) & 0x7FF) / 8;
		if (st.y > 128)
			st.y -= 256;
		st.y = -st.y;

		st.z = ((VertexCoord[i] >> 22) & 0x3FF) / 4;
		if (st.z > 128)
			st.z -= 256;

		//
		// Write Deus Ex vertex: x, y, z, and a dummy "0000"
		// For example, a Deus Ex vertex might look like
		// this: 3D00C4FFC3FF0000
		// The first 2 bytes are the X coordinate (will be
		// stored on disk in reversed format, so in this 
		// example, X is 0x003D), the second 2 bytes are 
		// the Y coordinate, the third 2 bytes are the Z 
		// coordinate, and the last 2 bytes are not used.
		//
		if (fwrite(&st, sizeof(st), 1, fp_write) != 1) {
			printf("\n\nCould not write vertex !\n");
			delete [] VertexCoord;
			return;
		}

		// Keep track of min/max values for display at end.  These
		// are for display only
		if (st.x < minX)
			minX = st.x;
		else if (st.x > maxX)
			maxX = st.x;

		if (st.y < minY)
			minY = st.y;
		else if (st.y > maxY)
			maxY = st.y;

		if (st.z < minZ)
			minZ = st.z;
		else if (st.z > maxZ)
			maxZ = st.z;

		// If the experimental "hi res" option is passed-in,
		// multiple the coords by 256 (use this in conjunction
		// with the 1/256 scaling factor in the .uc file)
		if (HighResExperiment) {

			st.x = st.x * 256.0F;
			st.y = st.y * 256.0F;
			st.z = st.z * 256.0F;

		}
	}

	fclose(fp_write);

	// If the user has passed-in the "-ms" option, we've got a lot more work to do.
	// We're going to have to flip each polygon (1st and 3rd vertex) as well as
	// each polygon's UV map (1st and 3rd coordinate).  All of this polygon stuff
	// is in the _d.3d file.
	if (FixMilkShape) {

		// Open up the "_d.3d" for writing (binary mode).
		fp_write = fopen( DataFileName, "wb");
		if (fp_write == 0) {
			printf("\n\nCan't open file %s for writing !\n", DataFileName);
			return;
		}

		// Write number of polygons to the _d.3d file
		fwrite(&NumPolys, 2, 1, fp_write);

		// Write number of vertices to the _d.3d file
		fwrite(&NumVertices, 2, 1, fp_write);

		// Write a block of 44 dummy bytes (this data isn't used by anything)
		// A long is 4 bytes, so write out 11 of 'em
		dummy_long = 0;
		for (i = 0; i < 11; i++)
			fwrite(&dummy_long, 4, 1, fp_write);

		// Write polys and poly-related flags, UV coords, and texture associations
		for (i = 0; i < NumPolys; i++) {

			// Here's where we're flipping the polys (notice that 1 and 3 are swapped)
			fwrite(&DeusExPolygon[i].PolyIdx3, 2, 1, fp_write);	// Index to 3rd polygon vertex
			fwrite(&DeusExPolygon[i].PolyIdx2, 2, 1, fp_write);	// Index to 2nd polygon vertex
			fwrite(&DeusExPolygon[i].PolyIdx1, 2, 1, fp_write);	// Index to 1st polygon vertex

			fwrite(&DeusExPolygon[i].TriangleType, 1, 1, fp_write);
			fwrite(&DeusExPolygon[i].Color, 1, 1, fp_write);

			// Here's where we're flipping the UV coords
			fwrite(&DeusExPolygon[i].Ucoord3, 1, 1, fp_write);	// 3rd U coord
			fwrite(&DeusExPolygon[i].Vcoord3, 1, 1, fp_write);	// 3rd V coord
			fwrite(&DeusExPolygon[i].Ucoord2, 1, 1, fp_write);	// 2nd U coord
			fwrite(&DeusExPolygon[i].Vcoord2, 1, 1, fp_write);	// 2nd V coord
			fwrite(&DeusExPolygon[i].Ucoord1, 1, 1, fp_write);	// 1st U coord
			fwrite(&DeusExPolygon[i].Vcoord1, 1, 1, fp_write);	// 1st V coord

			fwrite(&DeusExPolygon[i].TextureNum, 1, 1, fp_write);
			fwrite(&DeusExPolygon[i].dummy, 1, 1, fp_write);
		}

		fclose(fp_write);

		delete [] DeusExPolygon;
	}

	// Display min/max x, y, and z values (just an FYI to the user)
	printf("\n  X coordinate range: %d to %d", minX, maxX);
	printf("\n  Y coordinate range: %d to %d", minY, maxY);
	printf("\n  Z coordinate range: %d to %d", minZ, maxZ);

	// Success!
	printf("\n\n*** Conversion complete ***\n");

	// Set our memory free
	delete [] VertexCoord;
	delete DataFileName;

}
